# -*- coding: utf-8 -*-
# Author: Pranav Khade, Iowa State University
# Please read the project licence file for the Copyrights.

"""The 'packman.anm' module is a collection of anm methods and applications built on packman.molecule API.

Notes:
    * Current apps list: - packman.anm.ANM, packman.anm.RDANM
    * Cite the following paper:
        

Example:
    * Review the packman.bin.PACKMAN.py file for the app use.

Todo:
    * Add new features
    * Add citations in the note
    * Use ``sphinx.ext.todo`` extension
    * Clean the code
    * Redesign print statements
    * Add Tutorial Link

"""

from .predict_hinge import predict_hinge
from .anm import ANM