# -*- coding: utf-8 -*-
# Author: Pranav Khade, Iowa State University
# Please read the project licence file for the Copyrights.

"""The 'packman.bin' is a collection of binaries built on packman.molecule API.

Notes:
    * Current binary list: - PACKMAN.py : A program to predict the hinge on the molecule given the atoms and relevent parameters

Example:

    packman -h

    OR

    python3 -m packman -h

Todo:
    * Add new features
    * Use ``sphinx.ext.todo`` extension
    * Clean the code
    * Redesign print statements
    * Add Tutorial Link

"""

from .PACKMAN import main